/*****************************************************************************\
 *  src/common/job_options.h  - Extra job options
 *****************************************************************************
 *  Copyright (C) 2002 The Regents of the University of California.
 *  Produced at Lawrence Livermore National Laboratory (cf, DISCLAIMER).
 *  Written by Mark Grondona <grondona1@llnl.gov>.
 *  CODE-OCEC-09-009. All rights reserved.
 *
 *  This file is part of Slurm, a resource management program.
 *  For details, see <https://slurm.schedmd.com/>.
 *  Please also read the included file: DISCLAIMER.
 *
 *  Slurm is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  In addition, as a special exception, the copyright holders give permission
 *  to link the code of portions of this program with the OpenSSL library under
 *  certain conditions as described in each individual source file, and
 *  distribute linked combinations including the two. You must obey the GNU
 *  General Public License in all respects for all of the code used other than
 *  OpenSSL. If you modify file(s) with this exception, you may extend this
 *  exception to your version of the file(s), but you are not obligated to do
 *  so. If you do not wish to do so, delete this exception statement from your
 *  version.  If you delete this exception statement from all source files in
 *  the program, then also delete it here.
 *
 *  Slurm is distributed in the hope that it will be useful, but WITHOUT ANY
 *  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 *  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 *  details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with Slurm; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA.
\*****************************************************************************/

#ifndef _JOB_OPTIONS_H
#define _JOB_OPTIONS_H

#include "src/common/list.h"

struct job_option_info {
	int type;
	char *option;
	char *optarg;
};

/*
 *  Create generic job options container.
 */
extern list_t *job_options_create(void);

/*
 *  Append option of type `type' and its argument to job options
 */
extern int job_options_append(list_t *opts, int type, const char *opt,
			      const char *optarg);

/*
 *  Pack all accumulated options into buf
 */
extern int job_options_pack(list_t *opts, buf_t *buf);

/*
 *  Unpack options from buf into options container opts.
 */
extern int job_options_unpack(list_t *opts, buf_t *buf);

#endif /* !_JOB_OPTIONS_H */
